/* $Id$ */
/*
 * Frame structure:
 *
 * byte  code
 *    0  SOH
 *    1  frame type
 *    2  block length (upper byte)
 *    3  block length (lower byte)
 *    4  STX
 *    5  data body (n bytes)
 *  n+5  ETX or ETB
 *  n+6  check sum (upper byte)
 *  n+7  check sum (lower byte)
 *
 * block_length : STX + data_body + E + check_sum
 * check_sum    : SOH + type + block_length + STX + data_body + E
 */
#include <stdlib.h>
#include <string.h> /* memcpy() */
#include "common.h"
#include "codes.h"
#include "utils.h"
#include "frame.h"


struct frame {
  unsigned char *buffer; /* pointer to buffer */
  unsigned char *data;   /* pointer to data in buffer */
  size_t size;           /* buffer length */
  size_t length;         /* data length */
};

static int check_null(Frame *p);
static void frame_set_data_length(Frame *p, size_t len);
static void frame_set_type(Frame *p, unsigned char type);
static void frame_set_eof(Frame *p, unsigned char eof);
static void frame_set_checksum(Frame *p, int sum);
static int checksum(const unsigned char *data, size_t n);


/*
 * Creates a frame.
 *
 * in: size  the frame size.
 * return: the pointer to frame.
 */
Frame *frame_new(size_t size)
{
  Frame *p;
  unsigned char *buf;

  DEBUG_PRINT("frame_new() --- begin.\n");
  p = XMALLOC(Frame, 1);
  DEBUG_PRINT("p OK\n");
  buf = XCALLOC(unsigned char, size);
  DEBUG_PRINT("buf %d OK\n", size);
  *buf = SOH;
  *(buf+1) = TYPE_DATA;
  int_to_uchar2(buf+2, 4); /* block length */
  *(buf+4) = STX;
  *(buf+5) = ETX;
  int_to_uchar2(buf+6, 0); /* check sum */
  p->buffer = buf;
  p->size   = size;
  p->data   = buf+5;
  p->length = 0;
  DEBUG_PRINT("frame_new() --- end.\n");
  return p;
}

/*
 * Free a frame.
 *
 * in: p  the pointer to frame.
 */
void frame_delete(Frame *p)
{
  XFREE(p->buffer);
  XFREE(p);
}

/*
 * Checks consistence of frame data.
 *
 * in: p  the pointer to frame.
 * return: TRUE if data in buffer are consistence, otherwise FALSE.
 */
int frame_check(Frame *p)
{
  unsigned char *buf;
  size_t len;
  int csum;

  if (!check_null(p)) {
    return FALSE;
  }
  buf = p->buffer;
  if (*buf != SOH) {
    return FALSE;
  }
  len = uchar2_to_int(buf+2); /* block length = STX+data+E+csum */
  p->length = len-4; /* data length */
  if (*(buf+4) != STX) {
    return FALSE;
  }
  if ((*(buf+len+1) != ETX) && (*(buf+len+1) != ETB)) {
    return FALSE;
  }
  csum = uchar2_to_int(buf+len+2);
  if ((checksum(buf, len+2)^csum) != 0xffff) {
    return FALSE;
  } else {
    return TRUE;
  }
}

/*
 * Returns the frame length in buffer.
 *
 * in: p  the pointer to frame.
 * return: frame length.
 */
size_t frame_length(Frame *p)
{
  if (!check_null(p)) {
    return 0;
  }
  return p->length+8;
}

/*
 * Returns the capacity of frame buffer.
 *
 * in: p  the pointer to frame.
 * return: maxiumum frame size.
 */
size_t frame_size(Frame *p)
{
  if (!check_null(p)) {
    return 0;
  }
  return p->size;
}

/*
 * Returns a pointer to frame buffer.
 *
 * in: p  the pointer to frame.
 * return: a pointer to frame buffer.
 */
unsigned char *frame_buffer(Frame *p)
{
  if (!check_null(p)) {
    return NULL;
  }
  return p->buffer;
}

/*
 * Copies from data body.
 *
 * in: p     the pointer to frame.
 * out: dst  the pointer to memory area.
 * return: a pointer to dst.
 */
unsigned char *frame_data_copy(Frame *p, unsigned char *dst)
{
  size_t len;

  if (!check_null(p)) {
    return NULL;
  }
  len = p->length;
  return memcpy(dst, p->data, len);
}

/*
 * Gets block length in frame buffer.
 *
 * in: p  the pointer to frame.
 * return: block length.
 */
size_t frame_get_blocklength(Frame *p)
{
  unsigned char *buf;
  int len;

  if (!check_null(p)) {
    return 0;
  }
  buf = p->buffer;
  len = uchar2_to_int(buf+2);
  return len;
}

/*
 * Gets type in frame buffer.
 *
 * in: p  the pointer to frame.
 * return: type.
 */
unsigned char frame_get_type(Frame *p)
{
  if (!check_null(p)) {
    return -1;
  }
  return *(p->buffer+2);
}

/*
 * Gets charator of frame end in frame buffer.
 *
 * in: p  the pointer to frame.
 * return: code of frame end.
 */
int frame_series_code(Frame *p)
{
  unsigned char c;
  int rc;

  if (!check_null(p)) {
    return -1;
  }
  c = *(p->buffer+p->length+5);
  switch (c) {
  case ETX:
    rc = FRAME_SERIES_END;
    break;
  case ETB:
    rc = FRAME_SERIES_CONTINUE;
    break;
  default:
    rc = -1;
    break;
  }
  return rc;
}

/*
 * Sets data to frame buffer.
 *
 * in: p     the pointer to frame.
 *     type  the type.
 *     src   the type.
 *     len   the type.
 * return: setted frame length.
 */
size_t frame_set_data(Frame *p, unsigned char type, unsigned char *src, size_t len)
{
  unsigned char eof;
  size_t size;
  int sum;
  
  if (!check_null(p)) {
    return 0;
  }
  eof = ETX;
  size = frame_size(p);
  if (size < len+8) {
    len = size-8;
    eof = ETB;
  }
  frame_set_type(p, type);
  frame_set_data_length(p, len);
  memcpy(p->data, src, len);
  frame_set_eof(p, eof);
  sum = checksum(p->buffer, len+6)^0xffff;
  frame_set_checksum(p, sum);
  return len;
}

/* |----+----|----+----|----> inner function <----|----+----|----+----| */

static int check_null(Frame *p)
{
  if (p == NULL) {
    return FALSE;
  }
  if (p->buffer == NULL) {
    return FALSE;
  }
  return TRUE;
}

static void frame_set_data_length(Frame *p, size_t len)
{
  unsigned char *buf;

  p->length = len; /* data body length */
  buf = p->buffer;
  int_to_uchar2(buf+2, len+4); /* block length */
}

static void frame_set_type(Frame *p, unsigned char type)
{
  unsigned char *buf;

  buf = p->buffer;
  *(buf+1) = type;
}

static void frame_set_eof(Frame *p, unsigned char eof)
{
  unsigned char *buf;
  size_t len;

  buf = p->buffer;
  len = p->length;
  *(buf+len+5) = eof;
}

static void frame_set_checksum(Frame *p, int sum)
{
  unsigned char *buf;
  size_t len;

  buf = p->buffer;
  len = p->length;
  int_to_uchar2(buf+len+6, sum);
}

/*
 * 16bits check sum.
 */
static int checksum(const unsigned char *data, size_t n)
{
  int sum;

  sum = 0;
  while (n-- > 0) {
    sum += *(data++);
    sum &= 0xffff;
  }
  return sum;
}
