/* $Id$ */
/*
 * Mobile phone's file transfor program.
 */
#include <ctype.h> /* isalpha() isspace() */
#include <unistd.h> /* sleep() */
#include "common.h"
#include "phonefile.h"
#include "readline.h"
#include "strconv.h"

#define DEVICE_NAME_LEN_MAX  256
#define DEFAULT_DEVICE       "/dev/ttyS0"
#define DEFAULT_BAUDRATE     230400  /*115200;*/
#define DEFAULT_FRAMESIZE    8192  /*4096;*/

#define DEFAULT_PROMPT "mpsftp> "
#define NUL  '\0'
#define QUOT '\"'
#define FILENAME_LEN_MAX 256
#define PHONE_NUMBER_SIZE_MAX 16

typedef struct command Command;
typedef struct location Location;
typedef struct entry Entry;
typedef struct resource Resource;

struct command {
  struct command *next;
  int type;
  int argc;
  union {
    const char **argv;
    const char *string;
  } u;
};
enum commandType {
  TEST,
  SYSTEM_COMMAND,
  SET_PARAMS,
  OPEN,
  CLOSE,
  PHONE_NUMBER,
  MONITOR,
  OPERATION_OPEN,
  OPERATION_CLOSE,
  GET_FILE,
  GET_FILE_LIST,
  GET_FILE_INFO,
  PUT_FILE,
  SLEEP,
  EXIT,
  NOT_FOUND
};
struct location {
  int last_column;
};
struct entry {
  const char *key;
  int type;
};

static Entry keywords[] = {
#if DEBUG
  { "test", TEST },
  { "sleep", SLEEP },
  { "open", OPEN },
  { "close", CLOSE },
  { "op_open", OPERATION_OPEN },
  { "op_close", OPERATION_CLOSE },
  { "phone", PHONE_NUMBER },
  { "monitor", MONITOR },
#endif /* DEBUG */
  { "set", SET_PARAMS },
  { "ls", GET_FILE_LIST },
  { "ll", GET_FILE_INFO },
  { "get", GET_FILE },
  { "put", PUT_FILE },
  { "exit", EXIT },
  { NULL, NOT_FOUND }  /* sentinel */
};

struct resource {
  PhoneFile *fp;
  char device[DEVICE_NAME_LEN_MAX];
  int baudrate;
  int framesize;
};


static int execute(Command *cmd, Resource *res);
static int exec_set_resource(Resource *res, int argc, const char **argv);
static int exec_filelist(FILE *fout, PhoneFile *fp);
static int exec_fileinfo(FILE *fout, PhoneFile *p, const char *remote);
static void print_filelist(FILE *fout, FileList *list);
static void print_fileinfo(FILE *fout, FileInfo *info);

static Command *test_command(void);
static Command *command_template(int type, int argc, char **argv);

static Command *command_add(Command *cmd, int type, int argc, void *p);
static Command *command_new(int type, int argc, void *p);
static void command_delete(Command *cmd);
static void command_argv_free(Command *cmd);

static int parse_line(Command **cmd, const char *line);
static int get_ident(int c, Location *loc, const char *buf);
static char *get_string(int c, Location *loc, const char *buf);
static char *get_quoted_string(int c, Location *loc, const char *buf);
static int skip_space(Location *loc, const char *buf);
static int read_char(Location *loc, const char *buf);


int main(int argc, char *argv[])
{
  int result;
  const char *line;
  const char *prompt = DEFAULT_PROMPT;
  Command *cmd;
  Resource res = {
    NULL,
    DEFAULT_DEVICE,
    DEFAULT_BAUDRATE,
    DEFAULT_FRAMESIZE
  };

  strconv_set_code(0);
  while ((line = readLine(prompt)) != NULL) {
    if (parse_line(&cmd, line) == EXIT) {
      break;
    }
    result = execute(cmd, &res);
    DEBUG_PRINT("execute() returns: %d\n", result);
    command_delete(cmd);
  }
  return EXIT_SUCCESS;
}

/* ----- command interpretor ----- */

static int execute(Command *cmd, Resource *res)
{
  FILE *fout = stdout;
  char phone[PHONE_NUMBER_SIZE_MAX];
  char buf_remote[FILENAME_LEN_MAX];
  const char *remote, *local;
  int level, rc = FALSE;

  while (cmd != NULL) {
    switch (cmd->type) {
    case SYSTEM_COMMAND:
      DEBUG_PRINT("command = '%s'\n", cmd->u.string);
      rc = system(cmd->u.string);
      break;
    case SET_PARAMS:
      rc = exec_set_resource(res, cmd->argc, cmd->u.argv);
      break;
    case SLEEP:
      sleep((cmd->argc < 1) ? 1 : atoi(*(cmd->u.argv)));
      fprintf(fout, "sleep.\n");
      break;
    case OPEN:
      res->fp = phoneFile_open(res->device);
      rc = (res->fp == NULL) ? FALSE : TRUE;
      fprintf(fout, "connect: %s %s\n", res->device, rc ? "OK." : "failed.");
      break;
    case CLOSE:
      rc = phoneFile_close(res->fp);
      fprintf(fout, "close.\n");
      break;
    case PHONE_NUMBER:
      phoneFile_get_phone_number(res->fp, phone, PHONE_NUMBER_SIZE_MAX);
      fprintf(fout, "phone = %s\n", phone);
      rc = TRUE;
      break;
    case MONITOR:
      level = phoneFile_get_receiving_level(res->fp);
      if (level >= 0) {
        fprintf(fout, "RX level = %d [dB 10^-6V]\n", level);
        rc = TRUE;
      }
      break;
    case OPERATION_OPEN:
      rc = phoneFile_operation_open(res->fp, res->baudrate, res->framesize);
      break;
    case OPERATION_CLOSE:
      phoneFile_operation_close(res->fp);
      rc = TRUE;
      break;
    case GET_FILE:
      if (cmd->argc > 0) {
        remote = *(cmd->u.argv);
        local = (cmd->argc < 2) ? remote : *(cmd->u.argv+1);
        fprintf(fout, "local: %s remote: %s\n", local, remote);
        strconv_to_term(buf_remote, FILENAME_LEN_MAX, remote);
        rc = phoneFile_get_file(res->fp, buf_remote, local);
      }
      break;
    case GET_FILE_LIST:
      rc = exec_filelist(fout, res->fp);
      break;
    case GET_FILE_INFO:
      if (cmd->argc > 0) {
        remote = *(cmd->u.argv);
        strconv_to_term(buf_remote, FILENAME_LEN_MAX, remote);
        rc = exec_fileinfo(fout, res->fp, buf_remote);
      }
      break;
    case PUT_FILE:
      if (cmd->argc > 0) {
        local  = *(cmd->u.argv);
        remote = (cmd->argc < 2) ? local : *(cmd->u.argv+1);
        fprintf(fout, "local: %s remote: %s\n", local, remote);
        strconv_to_term(buf_remote, FILENAME_LEN_MAX, remote);
        rc = phoneFile_put_file(res->fp, buf_remote, local);
      }
      break;
    case NOT_FOUND:
    default:
      rc = FALSE;
      fprintf(fout, "?Invalid command\n");
      DEBUG_PRINT("command type = %d\n", cmd->type);
      break;
    }
    DEBUG_PRINT("rc = %d\n", rc);
    cmd = cmd->next;
  }
  return rc;
}

static int exec_set_resource(Resource *res, int argc, const char **argv)
{
  const char *s1, *s2;

  if (argc < 2) {
    return FALSE;
  }
  s1 = *argv;
  s2 = *(argv+1);
  if (strcmp(s1, "device") == 0) {
    strcpy(res->device, s2);
  } else if (strcmp(s1, "baudrate") == 0) {
    res->baudrate = atoi(s2);
  } else if (strcmp(s1, "framesize") == 0) {
    res->framesize = atoi(s2);
  }
  DEBUG_PRINT("device = %s, baudrate = %d, framesize = %d\n",
              res->device, res->baudrate, res->framesize);
  return TRUE;
}

static int exec_filelist(FILE *fout, PhoneFile *fp)
{
  FileList *list;
  int rc;
  size_t num, i;

  DEBUG_PRINT("-- begin.\n");
  rc = phoneFile_get_file_list(fp, &list);
  if (rc) {
    print_filelist(fout, list);
    num = fileList_size(list);
    for (i = 0; i < num; i++) {
      exec_fileinfo(fout, fp, fileList_get(list, i));
    }
    fileList_delete(list);
  }
  DEBUG_PRINT("-- end. rc = %d\n", rc);
  return rc;
}

static int exec_fileinfo(FILE *fout, PhoneFile *fp, const char *remote)
{
  FileInfo *info;
  int rc;

  rc = phoneFile_get_file_info(fp, remote, &info);
  if (rc) {
    print_fileinfo(fout, info);
    fileInfo_delete(info);
  }
  return rc;
}

static void print_filelist(FILE *fout, FileList *list)
{
  char buf[FILENAME_LEN_MAX];
  size_t num, i;

  num = fileList_size(list);
  for (i = 0; i < num; i++) {
    strconv_from_term(buf, FILENAME_LEN_MAX, fileList_get(list, i));
    fprintf(fout, "filelist[%d]: %s\n", i, buf);
  }
}

static void print_fileinfo(FILE *fout, FileInfo *info)
{
  char buf[FILENAME_LEN_MAX];
  struct tm *time;
  time_t t;

  strconv_from_term(buf, FILENAME_LEN_MAX, fileInfo_filename(info));
  t = fileInfo_time(info);
  time = localtime(&t);
  fprintf(fout, "filename:%s\n", buf);
  fprintf(fout, "?:0x%08x\n", fileInfo_unknown(info));
  fprintf(fout, "size:%d\n", fileInfo_size(info));
  fprintf(fout, "time:%d/%02d/%02d %02d:%02d:%02d\n",
          time->tm_year+1900,
          time->tm_mon+1,
          time->tm_mday,
          time->tm_hour,
          time->tm_min,
          time->tm_sec);
  fprintf(fout, "usec:%d\n", fileInfo_usec(info));
}

static Command *test_command(void)
{
  Command *cmd, *head;
  char **argv;

  head = cmd = command_new(OPEN, 0, NULL);
  argv = XMALLOC(char *, 1);
  *argv = xstrdup("2"); /* wait */
  cmd = command_add(cmd, SLEEP, 1, argv);
  cmd = command_add(cmd, PHONE_NUMBER, 0, NULL);
  cmd = command_add(cmd, OPERATION_OPEN, 0, NULL);
  cmd = command_add(cmd, GET_FILE_LIST, 0, NULL);
  argv = XMALLOC(char *, 1);
  *argv = xstrdup("040211_1520~01.jpg"); /* remote */
  cmd = command_add(cmd, GET_FILE_INFO, 1, argv);
  argv = XMALLOC(char *, 2);
  *argv = xstrdup("040211_1520~01.jpg"); /* remote */
  *(argv+1) = xstrdup("test.jpg"); /* local */
  cmd = command_add(cmd, GET_FILE, 2, argv);
  argv = XMALLOC(char *, 2);
  *argv = xstrdup("test2.jpg"); /* local */
  *(argv+1) = xstrdup("test2-remote.jpg"); /* remote */
  cmd = command_add(cmd, PUT_FILE, 2, argv);
  cmd = command_add(cmd, OPERATION_CLOSE, 0, NULL);
  cmd = command_add(cmd, CLOSE, 0, NULL);
  return head;
}


/*
 * template for command a transaction.
 */
static Command *command_template(int type, int argc, char **argv)
{
  Command *cmd, *head;
  char **pp;

  head = cmd = command_new(OPEN, 0, NULL);
  pp = XMALLOC(char *, 1);
  *pp = xstrdup("2"); /* wait */
  cmd = command_add(cmd, SLEEP, 1, pp);
  cmd = command_add(cmd, PHONE_NUMBER, 0, NULL);
  cmd = command_add(cmd, OPERATION_OPEN, 0, NULL);
  cmd = command_add(cmd, type, argc, argv);
  cmd = command_add(cmd, OPERATION_CLOSE, 0, NULL);
  cmd = command_add(cmd, CLOSE, 0, NULL);
  return head;
}

static Command *command_add(Command *cmd, int type, int argc, void *p)
{
  Command *next;

  next = command_new(type, argc, p);
  if (cmd != NULL) {
    cmd->next = next;
  }
  return next;
}

static Command *command_new(int type, int argc, void *p)
{
  Command *cmd;

  cmd = XMALLOC(Command, 1);
  cmd->next = NULL;
  cmd->type = type;
  cmd->argc = argc;
  if (argc > 0) {
    cmd->u.argv = (const char **)p;
  } else {
    cmd->u.string = (const char *)p;
  }
  return cmd;
}

static void command_delete(Command *cmd)
{
  Command *next;

  while (cmd != NULL) {
    command_argv_free(cmd);
    next = cmd->next;
    XFREE(cmd);
    cmd = next;
  }
}

static void command_argv_free(Command *cmd)
{
  char **p;
  int n;

  n = cmd->argc;
  if (n > 0) {
    p = (char **)cmd->u.argv;
    while (n-- > 0) {
      XFREE(*p);
      p++;
    }
    XFREE(cmd->u.argv);
  } else {
    XFREE(cmd->u.string);
  }
}

/* ----- line parser functions ----- */

/*
 * Parse line and make command node.
 */
static int parse_line(Command **cmd, const char *line)
{
  char **argv, **p, *s;
  struct location loc;
  size_t len;
  int c, id, argc, rc = NOT_FOUND;

  loc.last_column = 0;
  c = skip_space(&loc, line);
  DEBUG_PRINT("c = '%c'\n", c);
  if (c == '!') {
    /* shell command */
    len = strcspn(line+loc.last_column, "\n");
    s = NULL;
    if (len > 1) {
      s = XMALLOC(char, len+1);
      strncpy(s, line+loc.last_column, len);
      *(s+len) = '\0';
    }
    DEBUG_PRINT("command = %s\n", s);
    *cmd = command_new(SYSTEM_COMMAND, 0, s);
  } else if (isalpha(c) || (c == '_')) {
    /* identity */
    id = get_ident(c, &loc, line);
    DEBUG_PRINT("ident = %d\n", id);
    argc = 0;
    argv = XMALLOC(char *, 10);
    p = argv;
    while ((c = skip_space(&loc, line)) != NUL) {
      if (c == QUOT) {
        s = get_quoted_string(c, &loc, line);
      } else {
        s = get_string(c, &loc, line);
      }
      DEBUG_PRINT("string '%s'\n", s);
      *(p++) = s;
      if (argc++ > 10) {
        break;
      }
    }
    if (argc < 1) {
      XFREE(argv);
    }
    switch (id) {
    case TEST:
      *cmd = test_command();
      break;
    case GET_FILE_LIST:
    case GET_FILE_INFO:
    case GET_FILE:
    case PUT_FILE:
      *cmd = command_template(id, argc, argv);
      break;
    case OPERATION_OPEN:
      *cmd = command_new(PHONE_NUMBER, 0, NULL);
      command_add(*cmd, id, argc, argv);
      break;
    default:
      *cmd = command_new(id, argc, argv);
      break;
    }
  } else {
    *cmd = command_new(NOT_FOUND, 0, NULL);
  }
  rc = (*cmd)->type;
  return rc;
}

static int skip_space(Location *loc, const char *buf)
{
  int c;

  for (;;) {
    c = read_char(loc, buf);
    if (!isspace(c)) {
      break;
    }
  }
  return c;
}

static int get_ident(int c, Location *loc, const char *buf)
{
  Entry *entry;
  const char *string, *key;

  string = get_string(c, loc, buf);
  /* search keyword */
  entry = keywords;
  while ((key = entry->key) != NULL) {
    if (strcmp(string, key) == 0) {
      break;
    }
    entry++;
  }
  return entry->type;
}

static char *get_string(int c, Location *loc, const char *buf)
{
  char *string, *s, *tmp;
  size_t i, len;

  len = 1;
  s = XMALLOC(char, len);
  i = 0;
  do {
    if (isspace(c)) {
      break;
    }
    if (i >= len) {
      len += 1;
      tmp = XREALLOC(char, s, len);
      if (tmp != NULL) {
        s = tmp;
      }
    }
    s[i++] = (char)c;
  } while ((c = read_char(loc, buf)) != NUL);
  s[i++] = NUL;
  string = XMALLOC(char, i);
  strcpy(string, s);
  XFREE(s);
  return string;
}

static char *get_quoted_string(int c, Location *loc, const char *buf)
{
  char *string, *s, *tmp;
  size_t i, len;

  len = 1;
  s = XMALLOC(char, len);
  i = 0;
  while ((c = read_char(loc, buf)) != QUOT) {
    if (i >= len) {
      len += 1;
      tmp = XREALLOC(char, s, len);
      if (tmp != NULL) {
        s = tmp;
      }
    }
    s[i++] = (char)c;
  }
  s[i++] = NUL;
  string = XMALLOC(char, i);
  strcpy(string, s);
  XFREE(s);
  return string;
}

static int read_char(Location *loc, const char *buf)
{
  int c;

  c = *(buf+loc->last_column);
  if (c == NUL) {
    return c;
  }
  (loc->last_column)++;
  return c;
}
