/* $Id$ */
#include <stdio.h>
#include <stdlib.h>
#include <time.h> /* ctime() */
#include <termios.h> /* struct termios */
#include <unistd.h> /* sleep() */
#include "common.h"
#include "file.h"
#include "filelist.h"
#include "fileinfo.h"
#include "frame.h"
#include "framedata.h"
#include "strconv.h"
#include "readwrite.h"
#include "serial_term.h"
#include "phonefile.h"

#define RETRY_COUNT 2

struct phoneFile {
  int fd;                /* file descriptor */
  int framesize;         /* framesize */
  struct termios oldtio; /* terminal setting before */
};

static int negotiate_baudrate(PhoneFile *fp, int rate);
static int negotiate_framesize(PhoneFile *fp, int size);
static int recv_data(PhoneFile *fp, FrameData **frameData);
static int send_data_delete(PhoneFile *fp, FrameData *frameData);
static int send_data(PhoneFile *fp, FrameData *frameData);


/*
 * Open a phone file.
 */
PhoneFile *phoneFile_open(const char *path)
{
  PhoneFile *p;
  struct termios *oldtio;
  int fd;

  fd = open_device(path);
  if (fd < 0) {
    return NULL;
  }
  p = XMALLOC(PhoneFile, 1);
  oldtio = &(p->oldtio);
  term_initialize(fd, oldtio);
  p->fd = fd;
  p->framesize = 512;
  return p;
}

/*
 * Close a phone file.
 */
int phoneFile_close(PhoneFile *p)
{
  struct termios *oldtio;
  int fd;

  fd = p->fd;
  oldtio = &(p->oldtio);
  term_restore(fd, oldtio);
  close_device(fd);
  XFREE(p);
  return TRUE;
}

/*
 * Open operation.
 */
int phoneFile_operation_open(PhoneFile *fp, int baudrate, int framesize)
{
  FrameData *cmd;
  int rate, size;

  if (!open_extend_mode(fp->fd)) {
    return FALSE;
  }
  cmd = frameData_set_protocol_version_new();
  if (!send_data_delete(fp, cmd)) {
    return FALSE;
  }
  rate = negotiate_baudrate(fp, baudrate);
  DEBUG_PRINT("baudrate = %d -> %d (request %d)\n", 600, rate, baudrate);
  if (rate < 0) {
    return FALSE;
  }
  term_set_baudrate(fp->fd, rate);
  sleep(1); /* just wait */
  size = negotiate_framesize(fp, framesize);
  DEBUG_PRINT("framesize = %d -> %d (request %d)\n", fp->framesize, size, framesize);
  if (size < 0) {
    return FALSE;
  }
  fp->framesize = size;
  return TRUE;
}

/*
 * Close operation.
 */
void phoneFile_operation_close(PhoneFile *fp)
{
  close_extend_mode(fp->fd);
}

/*
 * Get file list.
 */
int phoneFile_get_file_list(PhoneFile *fp, FileList **list)
{
  FrameData *cmd, *data;

  DEBUG_PRINT("--- begin.\n");
  cmd = frameData_get_file_list_new();
  if (!send_data_delete(fp, cmd)) {
    return FALSE;
  }
  if (!recv_data(fp, &data)) {
    return FALSE;
  }
  *list = fileList_new(data->data, data->length);
  frameData_delete(data);
  DEBUG_PRINT("--- end.\n");
  return TRUE;
}

/*
 * Get the file information.
 */
int phoneFile_get_file_info(PhoneFile *fp, const char *remote, FileInfo **info)
{
  FrameData *cmd, *data;

  DEBUG_PRINT("--- begin.\n");
  cmd = frameData_set_file_new(remote);
  if (!send_data_delete(fp, cmd)) {
    return FALSE;
  }
  cmd = frameData_get_file_info_new();
  if (!send_data_delete(fp, cmd)) {
    return FALSE;
  }
  if (!recv_data(fp, &data)) {
    return FALSE;
  }
  *info = fileInfo_new(data->data);
  frameData_delete(data);
  DEBUG_PRINT("--- end.\n");
  return TRUE;
}

/*
 * Gets file.
 */
int phoneFile_get_file(PhoneFile *fp, const char *remote, const char *local)
{
  FrameData *cmd, *data;

  DEBUG_PRINT("--- begin.\n");
  cmd = frameData_get_file_new(remote);
  if (!send_data_delete(fp, cmd)) {
    return FALSE;
  }
  if (!recv_data(fp, &data)) {
    return FALSE;
  }
  if (!file_write(local, data->data, data->length)) {
    return FALSE;
  }
  frameData_delete(data);
  DEBUG_PRINT("--- end.\n");
  return TRUE;
}

/*
 * Puts file.
 */
int phoneFile_put_file(PhoneFile *fp, const char *remote, const char *local)
{
  FrameData *cmd, *data;
  unsigned char *buf;
  int size;

  DEBUG_PRINT("--- begin.\n");
  size = file_size(local);
  if (size < 0) {
    return FALSE;
  }
  cmd = frameData_put_file_new(remote, size);
  if (!send_data_delete(fp, cmd)) {
    return FALSE;
  }
  buf = XMALLOC(unsigned char, size);
  if (!file_read(local, buf)) {
    return FALSE;
  }
  data = frameData_data_new(buf, size);
  if (!send_data_delete(fp, data)) {
    return FALSE;
  }
  DEBUG_PRINT("--- end.\n");
  return TRUE;
}

/*
 * Deletes file.
 */
int phoneFile_delete_file(PhoneFile *fp, const char *remote)
{
  FrameData *cmd, *data;

  DEBUG_PRINT("--- begin.\n");
  cmd = frameData_put_file_new(remote, 0);
  if (!send_data_delete(fp, cmd)) {
    return FALSE;
  }
  data = frameData_data_new(NULL, 0);
  if (!send_data_delete(fp, data)) {
    return FALSE;
  }
  DEBUG_PRINT("--- end.\n");
  return TRUE;
}

/*
 * Gets telephone number string.
 */
void phoneFile_get_phone_number(PhoneFile *fp, char *buf, size_t n)
{
  get_own_number(fp->fd, buf, n);
}

/*
 * Gets receiving level.
 */
int phoneFile_get_receiving_level(PhoneFile *fp)
{
  return get_receiving_level(fp->fd);
}

/* ----------------------------------------------------------- */

static int negotiate_baudrate(PhoneFile *fp, int rate)
{
  FrameData *cmd;
  int rates[] = {
    230400, 115200, 57600, 38400, 19200,
    9600, 4800, 2400, 1200, 600, -1
  };
  int *p;

  p = rates;
  if (rate > 0) {
    while (*p > rate) {
      p++;
    }
  }
  while ((rate = *(p++)) > 0) {
    DEBUG_PRINT("baudrate:%d?\n", rate);
    cmd = frameData_set_baudrate_new(rate);
    if (send_data_delete(fp, cmd)) {
      DEBUG_PRINT("baudrate %d OK.\n", rate);
      break;
    }
  }
  return rate;
}

static int negotiate_framesize(PhoneFile *fp, int size)
{
  FrameData *cmd;
  int sizes[] = { 8192, 4096, 2048, 1024, 512, -1 };
  int *p;

  p = sizes;
  if (size > 0) {
    while (*p > size) {
      p++;
    }
  }
  while ((size = *(p++)) > 0) {
    DEBUG_PRINT("framesize:%d?\n", size);
    cmd = frameData_set_framesize_new(size);
    if (send_data_delete(fp, cmd)) {
      DEBUG_PRINT("framesize %d OK.\n", size);
      break;
    }
  }
  return size;
}

static int recv_data(PhoneFile *fp, FrameData **frameData)
{
  Frame *frame;
  unsigned char *buf, eof, type;
  size_t buf_len, offset, len;
  int fd, rc = FALSE;

  DEBUG_PRINT("recv_data() --- begin.\n");
  fd = fp->fd;
  frame = frame_new(fp->framesize);
  buf_len = fp->framesize;
  buf = XMALLOC(unsigned char, buf_len);
  offset = 0;
  while ((len = recv_frame(fd, frame, RETRY_COUNT)) > 0) {
    DEBUG_PRINT("received size = %d\n", len);
    eof = frame_series_code(frame);
    type = frame_get_type(frame);
    DEBUG_PRINT("received type = 0x%02x, eof = %d\n", type, eof);
    switch (eof) {
    case FRAME_SERIES_END:
    case FRAME_SERIES_CONTINUE:
      if ((offset+len) > buf_len) {
        buf_len += fp->framesize;
        buf = XREALLOC(unsigned char, buf, buf_len);
      }
      frame_data_copy(frame, buf+offset);
      offset += len;
      if (eof == FRAME_SERIES_END) {
        /* end of data */
        *frameData = frameData_new(type, buf, offset);
        DEBUG_PRINT("type = 0x%02x\n", (*frameData)->type);
        DEBUG_PRINT("offset = %d\n", offset);
        DEBUG_PRINT("frameData is ready.\n");
        rc = TRUE;
        goto end;
      }
      break;
    default:
      /* unknown code */
      rc = FALSE;
      goto end;
      break;
    }
  }
 end:
  XFREE(buf);
  frame_delete(frame);
  DEBUG_PRINT("recv_data() --- end. rc = %d\n", rc);
  return rc;
}

static int send_data_delete(PhoneFile *fp, FrameData *frameData)
{
  int rc;

  rc = send_data(fp, frameData);
  frameData_delete(frameData);
  return rc;
}

static int send_data(PhoneFile *fp, FrameData *frameData)
{
  Frame *frame;
  unsigned char *pdata;
  size_t len, wlen;
  int rc = FALSE;

  DEBUG_PRINT("send_data() --- begin.\n");
  frame = frame_new(fp->framesize);
  pdata = frameData->data;
  len = frameData->length;
  while (len > 0) {
    wlen = frame_set_data(frame, frameData->type, pdata, len);
    if (send_frame(fp->fd, frame, RETRY_COUNT)) {
      /* next block */
      pdata += wlen;
      len -= wlen;
      rc = TRUE;
    } else {
      /* error */
      rc = FALSE;
      break;
    }
  }
  frame_delete(frame);
  DEBUG_PRINT("send_data() --- end. rc = %d\n", rc);
  return rc;
}
#if 0
static int send_data(PhoneFile *fp, FrameData *frameData)
{
  Frame *frame;
  unsigned char *pdata;
  size_t len, wlen;
  int rc = FALSE;

  DEBUG_PRINT("send_data() --- begin.\n");
  frame = frame_new(fp->framesize);
  wlen = frame_set_data(frame, frameData->type, NULL, 0);
  if (send_frame(fp->fd, frame, RETRY_COUNT)) {
    /* OK */
    rc = TRUE;
  } else {
    /* error */
    rc = FALSE;
    break;
  }
  frame_delete(frame);
  DEBUG_PRINT("send_data() --- end. rc = %d\n", rc);
  return rc;
}
#endif
