/* $Id$ */
#include <fcntl.h>   /* open(), O_RDWR, O_NOCTTY */
#include <unistd.h>  /* read(), write(), close() */
#include "common.h"
#include "error.h"
#include "codes.h"
#include "frame.h"
#include "readwrite.h"

#define REQ_OWN_NUM 0x8e /* request own number */
#define END_OWN_NUM 0xaa /* end of own number */


static int recv_expect_bytes(int fd, const unsigned char *expect, size_t n);
static char number_to_char(char c);
static int to_level(int c);
static unsigned char read_byte(int fd);
static void read_bytes(int fd, unsigned char *buf, size_t n);
static void write_byte(int fd, unsigned char buf);
static void write_bytes(int fd, const unsigned char *buf, size_t n);
static void debug_out(FILE *fp, const char *s, const unsigned char *buf, size_t n);


int recv_frame(int fd, Frame *frame, int retry)
{
  unsigned char *buf;
  int rlen, size;

  DEBUG_PRINT("-- begin.\n");
  size = -1;
  buf = frame_buffer(frame);
  do {
    read_bytes(fd, buf, 4); /* SOH+type+block_length */
    rlen = frame_get_blocklength(frame);
    if (rlen+4 > frame_size(frame)) {
      DEBUG_PRINT("cannnot receive: blocklength = %d\n", rlen);
      break;
    }
    read_bytes(fd, buf+4, rlen); /* STX+data+(ETX|ETB)+checksum */
    size = rlen-4;
    if (!frame_check(frame)) {
      write_byte(fd, NAK);
      retry--;
    } else {
      write_byte(fd, ACK);
      retry = 0;
    }
  } while (retry > 0);
  DEBUG_PRINT("-- end.\n");
  return size;
}

int send_frame(int fd, Frame *frame, int retry)
{
  unsigned char *buf, c;
  size_t len;
  int rc;

  rc = FALSE;
  buf = frame_buffer(frame);
  len = frame_length(frame);
  do {
    write_bytes(fd, buf, len);
    c = read_byte(fd);
    switch (c) {
    case ACK:
      retry = 0;
      rc = TRUE;
      break;
    case NAK:
      retry--;
      break;
    case CAN:
      retry = 0;
      rc = FALSE;
      break;
    case BEL:
    default:
      break;
    }
  } while (retry > 0);
  return rc;
}


void get_own_number(int fd, char *buf, size_t n)
{
  unsigned char c, s = REQ_OWN_NUM;

  write_bytes(fd, &s, 1);
  while ((c = read_byte(fd)) != END_OWN_NUM) {
    if (to_level(c) >= 0) {
      /* not own number */
      continue;
    }
    if (n-- <= 0) {
      break;
    }
    *(buf++) = number_to_char(c);
  }
  *buf = '\0';
}

int get_receiving_level(int fd)
{
  int c;

  c = read_byte(fd);
  return to_level(c);
}

int open_extend_mode(int fd)
{
  unsigned char s1[] = { 0xd0, 0x01, ENQ };
  unsigned char r1[] = { 0xd0, 0x01, ACK };
  unsigned char s2[] = { ENQ };
  unsigned char r2[] = { ACK };

  write_bytes(fd, s1, 3);
  if (recv_expect_bytes(fd, r1, 3) < 3) {
    return FALSE;
  }
  write_bytes(fd, s2, 1);
  if (recv_expect_bytes(fd, r2, 1) < 1) {
    return FALSE;
  }
  return TRUE;
}

void close_extend_mode(int fd)
{
  write_byte(fd, EOT);
}


static int recv_expect_bytes(int fd, const unsigned char *expect, size_t n)
{
  unsigned char c;
  size_t i;

  i = 0;
  while (i < n) {
    c = read_byte(fd);
    if (c != *expect) {
      break;
    }
    expect++;
    i++;
  }
  return i;
}

/* ----------------------------------------------------------- */

int open_device(const char *path)
{
  int fd;

  fd = open(path, O_RDWR | O_NOCTTY);
  if (fd < 0) {
    error_fatal("Cannot open device");
  }
  return fd;
}

void close_device(int fd)
{
  if (close(fd)) {
    error_fatal("Cannot close device");
  }
}

static char number_to_char(char c)
{
  const char *dial = "D1234567890*#ABC"; /* dial number table */
  return *(dial+(c & 0x0f));
}

static int to_level(int c)
{
  int level = -1;

  if ((c & 0xf0) == 0xc0) {
    level = (c & 0x0f)*2;
  }
  return level;
}

static unsigned char read_byte(int fd)
{
  unsigned char buf;

  read_bytes(fd, &buf, sizeof(unsigned char));
  return buf;
}

static void read_bytes(int fd, unsigned char *buf, size_t n)
{
  ssize_t size;

  while (n > 0) {
    size = read(fd, buf, n);
    if (size < 0) {
      error_fatal("Cannot read from device");
    }
#if DEBUG
    debug_out(stdout, "read", buf, size);
#endif /* DEBUG */
    n -= size;
    buf += size;
  }
}

static void write_byte(int fd, unsigned char buf)
{
  write_bytes(fd, &buf, sizeof(unsigned char));
}

static void write_bytes(int fd, const unsigned char *buf, size_t n)
{
  ssize_t size;

  while (n > 0) {
    size = write(fd, buf, n);
    if (size < 0) {
      error_fatal("Cannot write to device");
    }
#if DEBUG
    debug_out(stdout, "write", buf, size);
#endif /* DEBUG */
    n -= size;
    buf += size;
  }
}

static void debug_out(FILE *fp, const char *s, const unsigned char *buf, size_t n)
{
  fprintf(fp, "%s (%d):", s, n);
  while (n-- > 0) {
    fprintf(fp, " 0x%02x", *buf);
    buf++;
  }
  fprintf(fp, "\n");
  fflush(fp);
}
